import logging
from pathlib import Path
from typing import List

import redis

from simon_says.config import ConfigLoader

logger = logging.getLogger(__name__)


class DataStore:
    """ Persistence class """

    def __init__(self, config_path: Path = None) -> None:
        self.cfg = ConfigLoader(config_path).config if config_path else ConfigLoader().config
        redis_host = self.cfg.get("data_store", "redis_host")
        redis_port = self.cfg.get("data_store", "redis_port")
        logger.debug("Instantiating Redis client at %s:%s", redis_host, redis_port)
        self._redis = redis.Redis(host=redis_host, port=int(redis_port), db=0, decode_responses=True)

    def add(self, key: str, value: str) -> None:
        """ Add a record """

        logger.debug("Adding key %s to db", key)
        self._redis.execute_command("SET", key, value)

    def delete(self, key: str) -> None:
        """ Delete a record """

        logger.debug("Deleting record %s", key)
        self._redis.execute_command("DEL", key)

    def get(self, key: str) -> str:
        """ Get AlarmEvent by UID """

        logger.debug("Getting key %s from store", key)
        return self._redis.execute_command("GET", key)

    def get_all_keys(self, pattern: str) -> List[str]:
        """ Get all keys matching the given pattern """

        logger.debug("Retrieving all keys matching %s from store", pattern)
        return self._redis.execute_command(f"KEYS {pattern}")



"""
ContactID codes from: http://www.voip-sip-sdk.com/attachments/583/contact_id.pdf
"""

EVENT_CATEGORIES = {
    "100": "Alarms",
    "200": "Supervisory",
    "300": "Troubles",
    "400": "Open/Close, Remote Access",
    "500": "Bypasses/Disables",
    "600": "Test/Misc",
}

CODES = {
    "100": {"name": "Medical", "type": "zone"},
    "101": {"name": "Personal Emergency", "type": "zone"},
    "102": {"name": "Fail to report in", "type": "zone"},
    "110": {"name": "Fire", "type": "zone"},
    "111": {"name": "Smoke", "type": "zone"},
    "112": {"name": "Combustion", "type": "zone"},
    "113": {"name": "Water flow", "type": "zone"},
    "114": {"name": "Heat", "type": "zone"},
    "115": {"name": "Pull Station", "type": "zone"},
    "116": {"name": "Duct", "type": "zone"},
    "117": {"name": "Flame", "type": "zone"},
    "118": {"name": "Near Alarm", "type": "zone"},
    "120": {"name": "Panic", "type": "zone"},
    "121": {"name": "Duress", "type": "user"},
    "122": {"name": "Silent", "type": "zone"},
    "123": {"name": "Audible", "type": "zone"},
    "124": {"name": "Duress – Access granted", "type": "zone"},
    "125": {"name": "Duress – Egress granted", "type": "zone"},
    "130": {"name": "Burglary", "type": "zone"},
    "131": {"name": "Perimeter", "type": "zone"},
    "132": {"name": "Interior", "type": "zone"},
    "133": {"name": "24 Hour (Safe)", "type": "zone"},
    "134": {"name": "Entry/Exit", "type": "zone"},
    "135": {"name": "Day/night", "type": "zone"},
    "136": {"name": "Outdoor", "type": "zone"},
    "137": {"name": "Tamper", "type": "zone"},
    "138": {"name": "Near alarm", "type": "zone"},
    "139": {"name": "Intrusion Verifier", "type": "zone"},
    "140": {"name": "General Alarm", "type": "zone"},
    "141": {"name": "Polling loop open", "type": "zone"},
    "142": {"name": "Polling loop short", "type": "zone"},
    "143": {"name": "Expansion module failure", "type": "zone"},
    "144": {"name": "Sensor tamper", "type": "zone"},
    "145": {"name": "Expansion module tamper", "type": "zone"},
    "146": {"name": "Silent Burglary", "type": "zone"},
    "147": {"name": "Sensor Supervision Failure", "type": "zone"},
    "150": {"name": "24 Hour Non-Burglary", "type": "zone"},
    "151": {"name": "Gas detected", "type": "zone"},
    "152": {"name": "Refrigeration", "type": "zone"},
    "153": {"name": "Loss of heat", "type": "zone"},
    "154": {"name": "Water Leakage", "type": "zone"},
    "155": {"name": "Foil Break", "type": "zone"},
    "156": {"name": "Day Trouble", "type": "zone"},
    "157": {"name": "Low bottled gas level", "type": "zone"},
    "158": {"name": "High temp", "type": "zone"},
    "159": {"name": "Low temp", "type": "zone"},
    "161": {"name": "Loss of air flow", "type": "zone"},
    "162": {"name": "Carbon Monoxide detected", "type": "zone"},
    "163": {"name": "Tank level", "type": "zone"},
    "200": {"name": "Fire Supervisory", "type": "zone"},
    "201": {"name": "Low water pressure", "type": "zone"},
    "202": {"name": "Low CO2", "type": "zone"},
    "203": {"name": "Gate valve sensor", "type": "zone"},
    "204": {"name": "Low water level", "type": "zone"},
    "205": {"name": "Pump activated", "type": "zone"},
    "206": {"name": "Pump failure", "type": "zone"},
    "300": {"name": "System Trouble", "type": "zone"},
    "301": {"name": "AC Loss", "type": "zone"},
    "302": {"name": "Low system battery", "type": "zone"},
    "303": {"name": "RAM Checksum bad", "type": "zone"},
    "304": {"name": "ROM checksum bad", "type": "zone"},
    "305": {"name": "System reset", "type": "zone"},
    "306": {"name": "Panel programming changed", "type": "zone"},
    "307": {"name": "Self-test failure", "type": "zone"},
    "308": {"name": "System shutdown", "type": "zone"},
    "309": {"name": "Battery test failure", "type": "zone"},
    "310": {"name": "Ground fault", "type": "zone"},
    "311": {"name": "Battery Missing/Dead", "type": "zone"},
    "312": {"name": "Power Supply Overcurrent", "type": "zone"},
    "313": {"name": "Engineer Reset", "type": "user"},
    "320": {"name": "Sounder/Relay", "type": "zone"},
    "321": {"name": "Bell 1", "type": "zone"},
    "322": {"name": "Bell 2", "type": "zone"},
    "323": {"name": "Alarm relay", "type": "zone"},
    "324": {"name": "Trouble relay", "type": "zone"},
    "325": {"name": "Reversing relay", "type": "zone"},
    "326": {"name": "Notification Appliance Ckt. # 3", "type": "zone"},
    "327": {"name": "Notification Appliance Ckt. #4", "type": "zone"},
    "330": {"name": "System Peripheral trouble", "type": "zone"},
    "331": {"name": "Polling loop open", "type": "zone"},
    "332": {"name": "Polling loop short", "type": "zone"},
    "333": {"name": "Expansion module failure", "type": "zone"},
    "334": {"name": "Repeater failure", "type": "zone"},
    "335": {"name": "Local printer out of paper", "type": "zone"},
    "336": {"name": "Local printer failure", "type": "zone"},
    "337": {"name": "Exp. Module DC Loss", "type": "zone"},
    "338": {"name": "Exp. Module Low Batt.", "type": "zone"},
    "339": {"name": "Exp. Module Reset", "type": "zone"},
    "341": {"name": "Exp. Module Tamper", "type": "zone"},
    "342": {"name": "Exp. Module AC Loss", "type": "zone"},
    "343": {"name": "Exp. Module self-test fail", "type": "zone"},
    "344": {"name": "RF Receiver Jam Detect", "type": "zone"},
    "350": {"name": "Communication trouble", "type": "zone"},
    "351": {"name": "Telco 1 fault", "type": "zone"},
    "352": {"name": "Telco 2 fault", "type": "zone"},
    "353": {"name": "Long Range Radio xmitter fault", "type": "zone"},
    "354": {"name": "Failure to communicate event", "type": "zone"},
    "355": {"name": "Loss of Radio supervision", "type": "zone"},
    "356": {"name": "Loss of central polling", "type": "zone"},
    "357": {"name": "Long Range Radio VSWR problem", "type": "zone"},
    "370": {"name": "Protection loop", "type": "zone"},
    "371": {"name": "Protection loop open", "type": "zone"},
    "372": {"name": "Protection loop short", "type": "zone"},
    "373": {"name": "Fire trouble", "type": "zone"},
    "374": {"name": "Exit error alarm (zone)", "type": "zone"},
    "375": {"name": "Panic zone trouble", "type": "zone"},
    "376": {"name": "Hold-up zone trouble", "type": "zone"},
    "377": {"name": "Swinger Trouble", "type": "zone"},
    "378": {"name": "Cross-zone Trouble", "type": "zone"},
    "380": {"name": "Sensor trouble", "type": "zone"},
    "381": {"name": "Loss of supervision - RF", "type": "zone"},
    "382": {"name": "Loss of supervision - RPM", "type": "zone"},
    "383": {"name": "Sensor tamper", "type": "zone"},
    "384": {"name": "RF low battery", "type": "zone"},
    "385": {"name": "Smoke detector Hi sensitivity", "type": "zone"},
    "386": {"name": "Smoke detector Low sensitivity", "type": "zone"},
    "387": {"name": "Intrusion detector Hi sensitivity", "type": "zone"},
    "388": {"name": "Intrusion detector Low sensitivity", "type": "zone"},
    "389": {"name": "Sensor self-test failure", "type": "zone"},
    "391": {"name": "Sensor Watch trouble", "type": "zone"},
    "392": {"name": "Drift Compensation Error", "type": "zone"},
    "393": {"name": "Maintenance Alert", "type": "zone"},
    "400": {"name": "Open/Close", "type": "user"},
    "401": {"name": "O/C by user", "type": "user"},
    "402": {"name": "Group O/C", "type": "user"},
    "403": {"name": "Automatic O/C", "type": "user"},
    "404": {"name": "Late to O/C", "type": "user"},
    "405": {"name": "Deferred O/C (Obsolete- do not use )", "type": "user"},
    "406": {"name": "Cancel", "type": "user"},
    "407": {"name": "Remote arm/disarm", "type": "user"},
    "408": {"name": "Quick arm", "type": "user"},
    "409": {"name": "Keyswitch O/C", "type": "user"},
    "411": {"name": "Callback request made", "type": "user"},
    "412": {"name": "Successful download/access", "type": "user"},
    "413": {"name": "Unsuccessful access", "type": "user"},
    "414": {"name": "System shutdown command received", "type": "user"},
    "415": {"name": "Dialer shutdown command received", "type": "user"},
    "416": {"name": "Successful Upload", "type": "zone"},
    "421": {"name": "Access denied", "type": "user"},
    "422": {"name": "Access report by user", "type": "user"},
    "423": {"name": "Forced Access", "type": "zone"},
    "424": {"name": "Egress Denied", "type": "user"},
    "425": {"name": "Egress Granted", "type": "user"},
    "426": {"name": "Access Door propped open", "type": "zone"},
    "427": {"name": "Access point Door Status Monitor trouble", "type": "zone"},
    "428": {"name": "Access point Request To Exit trouble", "type": "zone"},
    "429": {"name": "Access program mode entry", "type": "user"},
    "430": {"name": "Access program mode exit", "type": "user"},
    "431": {"name": "Access threat level change", "type": "user"},
    "432": {"name": "Access relay/trigger fail", "type": "zone"},
    "433": {"name": "Access RTE shunt", "type": "zone"},
    "434": {"name": "Access DSM shunt", "type": "zone"},
    "441": {"name": "Armed STAY", "type": "user"},
    "442": {"name": "Keyswitch Armed STAY", "type": "user"},
    "450": {"name": "Exception O/C", "type": "user"},
    "451": {"name": "Early O/C", "type": "user"},
    "452": {"name": "Late O/C", "type": "user"},
    "453": {"name": "Failed to Open", "type": "user"},
    "454": {"name": "Failed to Close", "type": "user"},
    "455": {"name": "Auto-arm Failed", "type": "user"},
    "456": {"name": "Partial Arm", "type": "user"},
    "457": {"name": "Exit Error (user)", "type": "user"},
    "458": {"name": "User on Premises", "type": "user"},
    "459": {"name": "Recent Close", "type": "user"},
    "461": {"name": "Wrong Code Entry", "type": "zone"},
    "462": {"name": "Legal Code Entry", "type": "user"},
    "463": {"name": "Re-arm after Alarm", "type": "user"},
    "464": {"name": "Auto-arm Time Extended", "type": "user"},
    "465": {"name": "Panic Alarm Reset", "type": "zone"},
    "466": {"name": "Service On/Off Premises", "type": "user"},
    "501": {"name": "Access reader disable", "type": "zone"},
    "520": {"name": "Sounder/Relay Disable", "type": "zone"},
    "521": {"name": "Bell 1 disable", "type": "zone"},
    "522": {"name": "Bell 2 disable", "type": "zone"},
    "523": {"name": "Alarm relay disable", "type": "zone"},
    "524": {"name": "Trouble relay disable", "type": "zone"},
    "525": {"name": "Reversing relay disable", "type": "zone"},
    "526": {"name": "Notification Appliance Ckt. # 3 disable", "type": "zone"},
    "527": {"name": "Notification Appliance Ckt. # 4 disable", "type": "zone"},
    "531": {"name": "Module Added", "type": "zone"},
    "532": {"name": "Module Removed", "type": "zone"},
    "551": {"name": "Dialer disabled", "type": "zone"},
    "552": {"name": "Radio transmitter disabled", "type": "zone"},
    "553": {"name": "Remote Upload/Download disabled", "type": "zone"},
    "570": {"name": "Zone/Sensor bypass", "type": "zone"},
    "571": {"name": "Fire bypass", "type": "zone"},
    "572": {"name": "24 Hour zone bypass", "type": "zone"},
    "573": {"name": "Burg. Bypass", "type": "zone"},
    "574": {"name": "Group bypass", "type": "user"},
    "575": {"name": "Swinger bypass", "type": "zone"},
    "576": {"name": "Access zone shunt", "type": "zone"},
    "577": {"name": "Access point bypass", "type": "zone"},
    "601": {"name": "Manual trigger test report", "type": "zone"},
    "602": {"name": "Periodic test report", "type": "zone"},
    "603": {"name": "Periodic RF transmission", "type": "zone"},
    "604": {"name": "Fire test", "type": "user"},
    "605": {"name": "Status report to follow", "type": "zone"},
    "606": {"name": "Listen-in to follow", "type": "zone"},
    "607": {"name": "Walk test mode", "type": "user"},
    "608": {"name": "Periodic test - System Trouble Present", "type": "zone"},
    "609": {"name": "Video Xmitter active", "type": "zone"},
    "611": {"name": "Point tested OK", "type": "zone"},
    "612": {"name": "Point not tested", "type": "zone"},
    "613": {"name": "Intrusion Zone Walk Tested", "type": "zone"},
    "614": {"name": "Fire Zone Walk Tested", "type": "zone"},
    "615": {"name": "Panic Zone Walk Tested", "type": "zone"},
    "616": {"name": "Service Request", "type": "zone"},
    "621": {"name": "Event Log reset", "type": "zone"},
    "622": {"name": "Event Log 50% full", "type": "zone"},
    "623": {"name": "Event Log 90% full", "type": "zone"},
    "624": {"name": "Event Log overflow", "type": "zone"},
    "625": {"name": "Time/Date reset", "type": "user"},
    "626": {"name": "Time/Date inaccurate", "type": "zone"},
    "627": {"name": "Program mode entry", "type": "zone"},
    "628": {"name": "Program mode exit", "type": "zone"},
    "629": {"name": "32 Hour Event log marker", "type": "zone"},
    "630": {"name": "Schedule change", "type": "zone"},
    "631": {"name": "Exception schedule change", "type": "zone"},
    "632": {"name": "Access schedule change", "type": "zone"},
    "641": {"name": "Senior Watch Trouble", "type": "zone"},
    "642": {"name": "Latch-key Supervision", "type": "user"},
    "651": {"name": "Reserved for Ademco Use", "type": "zone"},
    "652": {"name": "Reserved for Ademco Use", "type": "user"},
    "653": {"name": "Reserved for Ademco Use", "type": "user"},
    "654": {"name": "System Inactivity", "type": "zone"},
}



import json
import logging
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List

from pydantic import BaseModel

from simon_says.config import ConfigLoader

logger = logging.getLogger(__name__)


class SensorState(Enum):
    OPEN = "open"
    CLOSED = "closed"
    BYPASSED = "bypassed"


class Sensor(BaseModel):
    """ An alarm Sensor """

    number: int
    name: str
    state: SensorState = SensorState.CLOSED

    def to_dict(self) -> Dict[str, Any]:
        """ Convert to Dict """
        res = self.__dict__
        res["state"] = self.state.value
        return res

    def to_json(self) -> str:
        """ Convert to JSON """
        return json.dumps(self.to_dict())


class Sensors:
    """
    A collection of Sensor objects.
    These correspond to "zones" in the Ademco nomenclature.
    """

    def __init__(self, config_path: Path = None) -> None:
        self._sensors_by_number = {}
        self.cfg = ConfigLoader(config_path).config if config_path else ConfigLoader().config
        self._load_from_config()

    def add(self, sensor: Sensor) -> None:
        """ Add a sensor to the collection """
        if sensor.number in self._sensors_by_number:
            raise ValueError("Sensor number %s already exists")

        self._sensors_by_number[sensor.number] = sensor

    def by_number(self, number: int) -> Sensor:
        """ Get sensor given its number """
        return self._sensors_by_number[number]

    def get_all_sensors(self) -> List[Sensor]:
        """ Get all sensors in the collection """

        return list(self._sensors_by_number.values())

    def all_as_json(self) -> str:
        """ Return all sensors as JSON """

        return json.dumps([s.to_dict() for s in self.get_all_sensors()])

    def clear_all(self) -> None:
        """ Clear all sensors (set to CLOSED state) """

        logger.debug("Clearing all sensors")
        for sensor in self.get_all_sensors():
            sensor.state = SensorState.CLOSED

    def _load_from_config(self) -> None:
        """ Load sensors into collection from config data """
        for number, name in self.cfg["sensors"].items():
            self.add(Sensor(number=int(number), name=name))



import pytest

from simon_says.helpers import redis_present

test_data = {
    "test:1": "foo",
    "test:2": "bar",
    "other": "baz",
}


@pytest.mark.skipif(not redis_present(), reason="redis not present")
def test_db_crud(test_db):
    for k, v in test_data.items():
        test_db.add(k, v)
        assert test_db.get(k) == v

    test_keys = test_db.get_all_keys("test:*")
    assert len(test_keys) == 2

    for k, v in test_data.items():
        test_db.delete(k)

    all_keys = test_db.get_all_keys("test:*")
    assert len(all_keys) == 0



from typing import Any, Dict, List

import requests

# This is both the connect and read timeout values
# Notice that this does not apply to the total length of the request
# See: https://requests.readthedocs.io/en/latest/user/advanced/#timeouts
DEFAULT_TIMEOUT = 10


class Client(object):
    def __init__(self, url: str):
        self._url = url
        self._session = requests.Session()

    def arm_home(self, timeout: int = DEFAULT_TIMEOUT) -> str:
        r = self._session.post(f"{self._url}/control", json={"action": "arm_home"}, timeout=timeout)
        if r.status_code == 202:
            return r.text
        else:
            raise RuntimeError(f"Error code: {r.status_code}, content: {r.text}")

    def arm_away(self, timeout: int = DEFAULT_TIMEOUT) -> str:
        r = self._session.post(f"{self._url}/control", json={"action": "arm_away"}, timeout=timeout)
        if r.status_code == 202:
            return r.text
        else:
            raise RuntimeError(f"Error code: {r.status_code}, content: {r.text}")

    def disarm(self, timeout: int = DEFAULT_TIMEOUT) -> str:
        r = self._session.post(f"{self._url}/control", json={"action": "disarm"}, timeout=timeout)
        if r.status_code == 202:
            return r.text
        else:
            raise RuntimeError(f"Error code: {r.status_code}, content: {r.text}")

    def add_event(self, data: Dict, timeout: int = DEFAULT_TIMEOUT) -> str:
        """ Add a single event """
        r = self._session.post(f"{self._url}/events", json=data, timeout=timeout)
        if r.status_code == 201:
            return r.text
        else:
            raise RuntimeError(f"Error code: {r.status_code}, content: {r.text}")

    def get_events(self, timeout: int = DEFAULT_TIMEOUT) -> List[Dict[str, Any]]:
        """ Get all events in queue """
        r = self._session.get(f"{self._url}/events", timeout=timeout)
        if r.status_code == 200:
            data = r.json()
            return data
        else:
            raise RuntimeError(f"Error code: {r.status_code}, content: {r.text}")

    def get_event(self, uid: str, timeout: int = DEFAULT_TIMEOUT) -> Dict[str, Any]:
        """ Get a single event given its UID """
        r = self._session.get(f"{self._url}/events/{uid}", timeout=timeout)
        if r.status_code == 200:
            data = r.json()
            return data
        else:
            raise RuntimeError(f"Error code: {r.status_code}, content: {r.text}")



import logging
from configparser import ConfigParser
from pathlib import Path

DEFAULT_CONFIG_PATH = Path("/etc/simon_says.ini")

logger = logging.getLogger(__name__)

DEFAULTS = {
    "data_store": {
        "redis_host": "localhost",
        "redis_port": 6379,
    },
    "events": {
        # Default directories to read files from and move them to on the Asterisk instance
        "src_dir": "/var/spool/asterisk/alarm_events",
        "dst_dir": "/var/spool/asterisk/alarm_events_processed",
    },
    "control": {
        # Default alarm access code
        "access_code": "1234",
        # SIP extension that will receive the commands via Asterisk
        "extension": "100",
        # How long to wait for the alarm to answer
        # Unfortunately it has to ring 10 times, which is roughly 1 minute.
        "wait_time": 65,
        # Wait 10 seconds between retries
        "retry_time": 10,
        # Retry this many times
        "max_retries": 2,
        # Default user that will own the call files
        "asterisk_user": "asterisk",
        # Default spool directory
        "spool_dir": "/var/spool/asterisk/outgoing",
    },
    # Default sensor (zone) names.
    "sensors": {"0": "nothing"},
}


class ConfigLoader:
    def __init__(self, cfg_path: Path = DEFAULT_CONFIG_PATH) -> None:
        self.config = ConfigParser()
        self.load_defaults()
        self.load_from_file(cfg_path)

    def load_defaults(self) -> None:
        self.config.read_dict(DEFAULTS)

    def load_from_file(self, cfg_path) -> None:
        """ Load config file """

        if cfg_path.is_file():
            self.config.read(cfg_path)
        else:
            logger.warning("No configuration file found at %s. Using defaults", cfg_path)



import logging
from pathlib import Path

from pycall import Application, Call, CallFile

from simon_says.config import ConfigLoader

# Map relevant actions to DTMF sequences
# See user manual at https://static.interlogix.com/library/466-2266_rev_f.pdf
ACTION_TO_DTMF = {
    "disarm": ["1"],
    "arm_doors_and_windows": ["2"],
    "arm_motion_sensors": ["3"],
    "arm_doors_and_windows_no_delay": ["2", "2"],
    "arm_motion_sensors_with_latchkey": ["3", "3"],
    "arm_doors_and_windows_and_motion_sensors": ["2", "3"],
    "arm_doors_and_windows_with_no_entry_delay_and_motion_sensors_with_latchkey": ["2", "2", "3", "3"],
    "terminate": ["9"],
}

logger = logging.getLogger(__name__)


class Controller:
    """ SimonXT controller class """

    def __init__(
        self,
        config_path: Path = None,
        access_code: str = None,
        extension: str = None,
        wait_time: int = None,
        retry_time: int = None,
        max_retries: int = None,
        asterisk_user: str = None,
        spool_dir: Path = None,
    ) -> None:
        self.cfg = ConfigLoader(config_path).config if config_path else ConfigLoader().config
        self._state_db_key = "armed_state"
        self.access_code = access_code or self.cfg.get("control", "access_code")
        self.extension = extension or self.cfg.get("control", "extension")
        self.wait_time = wait_time or int(self.cfg.get("control", "wait_time"))
        self.retry_time = retry_time or int(self.cfg.get("control", "retry_time"))
        self.max_retries = max_retries or int(self.cfg.get("control", "max_retries"))
        self.asterisk_user = asterisk_user or self.cfg.get("control", "asterisk_user")
        self.spool_dir = spool_dir or Path(self.cfg.get("control", "spool_dir"))

        if not self.spool_dir.is_dir():
            raise ValueError(f"spool_dir {self.spool_dir} is not a valid directory")

    def _build_dtmf_sequence(self, action: str) -> str:
        """
        Build DTMF tone sequence to send to alarm
        """

        if action not in ACTION_TO_DTMF:
            raise ValueError(f"Invalid action: {action}")

        # "w" means wait a half second
        # For more details, see https://wiki.asterisk.org/wiki/display/AST/Application_SendDTMF

        # Wait before sending the access code
        sections = ["w", self.access_code]

        # Add requested action
        sections.extend(ACTION_TO_DTMF[action])

        # Hang up
        sections.extend(ACTION_TO_DTMF["terminate"])

        # Join all tones with a half-second in between them
        result = "w".join(sections)

        return result

    def send_command(self, action: str) -> None:
        """ Send control sequence via Asterisk call file """

        call = Call(
            f"SIP/{self.extension}", wait_time=self.wait_time, retry_time=self.retry_time, max_retries=self.max_retries
        )

        seq = self._build_dtmf_sequence(action)

        logger.debug("Sending action '%s' (DTMF: '%s') to alarm", action, seq)

        action = Application("SendDTMF", seq)

        callfile_args = {"archive": True, "spool_dir": self.spool_dir}
        if self.asterisk_user:
            callfile_args["user"] = self.asterisk_user

        c = CallFile(call, action, **callfile_args)
        c.spool()

    def disarm(self) -> None:
        """ Disarm """

        self.send_command("disarm")

    def arm_home(self) -> None:
        """ Arm while at home """

        self.send_command("arm_doors_and_windows_no_delay")

    def arm_away(self) -> None:
        """ Arm when going away """

        self.send_command("arm_doors_and_windows_and_motion_sensors")



import json

import pytest

from simon_says.sensors import Sensor, Sensors, SensorState


@pytest.fixture
def test_sensors(test_config_path):
    return Sensors(config_path=test_config_path)


@pytest.fixture
def all_test_sensors(test_sensors):
    all_sensors = test_sensors.get_all_sensors()
    assert len(all_sensors) == 5
    return all_sensors


def test_sensor_ops(test_sensors, all_test_sensors):
    for sensor in all_test_sensors:
        assert isinstance(sensor, Sensor)
        assert test_sensors.by_number(sensor.number) == sensor
        assert sensor.state == SensorState.CLOSED
        sensor.state = SensorState.OPEN
        assert sensor.state == SensorState.OPEN

    test_sensors.clear_all()
    for sensor in all_test_sensors:
        assert sensor.state == SensorState.CLOSED


def test_sensors_all_as_json(test_sensors):
    j_str = test_sensors.all_as_json()
    data = json.loads(j_str)
    assert len(data) == 5





def test_build_dtmf(test_controller):
    dtmf = test_controller._build_dtmf_sequence("arm_doors_and_windows_no_delay")
    assert dtmf == "ww1234w2w2w9"


def test_send_disarm_command(test_controller, tmp_path):
    test_controller.send_command("disarm")
    call_file = next(tmp_path.iterdir())
    lines = call_file.read_text().splitlines()
    assert lines[0] == "Channel: SIP/100"
    assert lines[1] == "WaitTime: 65"
    assert lines[2] == "RetryTime: 10"
    assert lines[3] == "Maxretries: 2"
    assert lines[4] == "Application: SendDTMF"
    assert lines[5] == "Data: ww1234w1w9"
    assert lines[6] == "Archive: yes"
    call_file.unlink()


def test_arm_doors_and_windows_no_delay(test_controller, tmp_path):
    test_controller.send_command("arm_doors_and_windows_no_delay")
    call_file = next(tmp_path.iterdir())
    lines = call_file.read_text().splitlines()
    assert lines[5] == "Data: ww1234w2w2w9"
    call_file.unlink()



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

__version__ = None
exec(open("simon_says/version.py").read())

setup(
    name="simon_says",
    version=__version__,
    description="GE/Interlogix Simon XT Alarm interface library and API",
    long_description="Interact with the Simon XT Alarm system",
    author="Carlos Vicente",
    author_email="cvicente@gmail.com",
    url="https://github.com/cvicente/simon_says",
    license="GPL v3.0",
    classifiers=(
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
    ),
    packages=find_packages(exclude=["tests", "tests.*"]),
    # required for PEP-561 compatible typed packages.
    zip_safe=False,
    setup_requires=["setuptools", "pytest"],
    # pip install consults this list by specifying . in requirements.txt
    install_requires=[
        "configparser",
        "falcon",
        "gunicorn",
        "httpie",
        "pycall",
        "pydantic",
        "pyyaml",
        "redis",
        "requests",
    ],
    extras_require={
        "dev": ["mock", "pytest", "pytest-localserver", "pytest-mock", "tox"],
        "lint": ["black", "flake8", "isort"],
    },
    scripts=["bin/simon_event_handler", "bin/simon_event_handler"],
)



import socket
from contextlib import closing

from simon_says.config import ConfigLoader

config = ConfigLoader().config


def port_open(host, port) -> bool:
    """ Check if TCP port is open """
    try:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            if sock.connect_ex((host, port)) == 0:
                return True
            else:
                return False
    except socket.gaierror:
        return False


def redis_present() -> bool:
    """ Check if Redis service is present """
    return port_open(config.get("data_store", "redis_host"), int(config.get("data_store", "redis_port")))



import falcon
import pytest
from falcon import testing

from simon_says.app import create_app
from simon_says.events import EventStore
from simon_says.helpers import redis_present

pytestmark = pytest.mark.skipif(not redis_present(), reason="redis not present")


@pytest.fixture
def client(test_controller):
    app = create_app(controller=test_controller)
    return testing.TestClient(app)


def test_get_version(client):
    response = client.simulate_get("/version")
    result = response.json
    assert result["version"]


def test_post_and_get_events(client, test_parsed_events, test_db):
    store = EventStore(db=test_db)

    for rec in test_parsed_events:
        # Delete the event first if it already exists
        if store.get(rec["uid"]):
            store.delete(rec["uid"])

        res = client.simulate_post("/events", json=rec)
        assert res.status == falcon.HTTP_CREATED

    response = client.simulate_get("/events")
    assert response.status == falcon.HTTP_OK

    result = response.json
    assert len(result) == 2

    uid = "12abcd"
    e1 = result[0]
    assert e1["uid"] == uid

    response = client.simulate_get(f"/events/{uid}")
    result = response.json
    assert result["uid"] == uid


def test_controller_disarm(client, tmp_path):
    data = {"action": "disarm"}
    resp = client.simulate_post("/control", json=data)
    assert resp.status == falcon.HTTP_ACCEPTED

    call_file = next(tmp_path.iterdir())
    lines = call_file.read_text().splitlines()
    assert lines[5] == "Data: ww1234w1w9"
    call_file.unlink()


def test_get_sensors(client):
    response = client.simulate_get("/sensors")
    result = response.json
    assert len(result) == 1


def test_get_one_sensor(client):
    response = client.simulate_get("/sensors/0")
    result = response.json
    assert result["name"] == "nothing"
    assert result["state"] == "closed"



import logging
import os
import sys
from typing import List

DEFAULT_LOGLEVEL = "INFO"


def configure_logging(log_level: str = None, handlers: List[logging.Handler] = None) -> None:
    """
    Configure logging
    """

    # Get root logger
    root = logging.getLogger()

    if log_level:
        # Set the level passed to this function
        root.setLevel(log_level)
    else:
        # Check if ENV log level is set, otherwise use default
        log_level = os.environ.get("SIMON_SAYS_LOGLEVEL") or DEFAULT_LOGLEVEL
        root.setLevel(log_level)
    if handlers:
        # Use the handlers passed to this function
        root.handlers = handlers
    else:
        # Or use the default handler
        root.addHandler(logging.StreamHandler(sys.stdout))



import configparser
import datetime
import json
import logging
import re
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from simon_says.ademco import CODES, EVENT_CATEGORIES
from simon_says.config import ConfigLoader
from simon_says.db import DataStore

logger = logging.getLogger(__name__)


class AlarmEvent(BaseModel):
    """ Represents an alarm event """

    uid: str
    timestamp: int
    extension: str
    account: int
    msg_type: int
    qualifier: int
    code: int
    code_description: str
    category: str
    partition: int
    sensor: Optional[int]
    sensor_name: Optional[str]
    user: Optional[int]
    checksum: int
    status: Optional[str]

    def to_dict(self) -> Dict[str, Any]:
        """ Convert to Dict """
        return self.__dict__

    def to_json(self) -> str:
        """ Convert event to JSON """
        return json.dumps(self.to_dict())


class EventStore:
    """ A store of alarm events """

    def __init__(self, db: DataStore) -> None:
        self._namespace = "event"
        self._db = db

    def add(self, event: AlarmEvent) -> None:
        """ Add an event """

        if self._db.get(self.obj_key(event.uid)):
            raise ValueError(f"Event with uid {event.uid} already exists")

        logger.debug("Adding AlarmEvent %s to store", event.uid)
        self._db.add(self.obj_key(event.uid), event.to_json())

    def delete(self, uid: str) -> None:
        """ Delete an event given its UID """

        logger.debug("Deleting event %s", uid)
        self._db.delete(self.obj_key(uid))

    def get(self, uid: str) -> Optional[AlarmEvent]:
        """ Get AlarmEvent by UID """

        logger.debug("Getting event %s from store", uid)

        j_str = self._db.get(self.obj_key(uid))
        if not j_str:
            return None

        obj_data = json.loads(j_str)
        return AlarmEvent(**obj_data)

    def get_all_keys(self) -> List[str]:
        """ Get all keys in our namespace """

        logger.debug("Retrieving all %s keys from store", self._namespace)
        return self._db.get_all_keys(f"{self._namespace}*")

    def obj_key(self, uid: str) -> str:
        """ Return the key string used to store and retrieve event objects """

        return f"{self._namespace}:{uid}"

    def get_events(self) -> List[AlarmEvent]:
        """ Get all events in store, in chronological order """

        logger.debug("Retrieving all events from store")
        res = []
        for key in self.get_all_keys():
            obj_data = json.loads(self._db.get(key))
            res.append(AlarmEvent(**obj_data))

        return sorted(res, key=lambda x: x.timestamp)

    def events_as_json(self) -> str:
        """ Get all events as a list, in JSON format """

        logger.debug("Retrieving all events, in JSON format")
        return json.dumps([e.to_dict() for e in self.get_events()])


class EventParser:
    """
    Parse Asterisk's AlarmReceiver events
    """

    def __init__(
        self,
        config_path: Path = None,
        src_dir: Path = None,
        dst_dir: Path = None,
        move_files: bool = True,
    ) -> None:

        self.cfg = ConfigLoader(config_path).config if config_path else ConfigLoader().config
        self.src_dir = src_dir or Path(self.cfg.get("events", "src_dir"))
        self.dst_dir = dst_dir or Path(self.cfg.get("events", "dst_dir"))

        for p in (self.src_dir, self.dst_dir):
            if not p.is_dir():
                raise RuntimeError(f"Required directory {p} does not exist")

        self.move_files = move_files

    def parse_file(self, path: Path) -> Optional[Dict[str, Any]]:
        """
        Parse an event file
        See: https://www.voip-info.org/asterisk-cmd-alarmreceiver/

        Notice that this code expects Asterisk's EventHandler config to be set with:
            logindividualevents = yes
        """

        logger.debug("Parsing event file at %s", path)
        with path.open("r") as f:
            for line in f:
                line = line.strip()

                # Verify Protocol
                proto_match = re.match(r"PROTOCOL=(.*)$", line)
                if proto_match:
                    proto_value = proto_match.group(1)
                    if proto_value != "ADEMCO_CONTACT_ID":
                        logger.warning("Invalid protocol. Skipping file {p}")
                        break

                uid = self._get_uid_from_filename(path.name)

                # Get caller extension
                x_match = re.match(r"CALLINGFROM=(.*)$", line)
                if x_match:
                    extension = x_match.group(1)

                # Get Timestamp
                t_match = re.match(r"TIMESTAMP=(.*)$", line)
                if t_match:
                    timestamp_str = t_match.group(1)
                    timestamp = self._parse_timestamp_str(timestamp_str)

                # Get event info
                fields = re.findall(r"^(\d{4})(\d{2})(\d)(\d{3})(\d{2})(\d{3})(\d)", line)
                if fields:
                    logger.debug("Event line found: %s", line)

                    account, msg_type, qualifier, code, partition, sensor_or_user, checksum = fields[0]

                    event_data = {
                        "uid": uid,
                        "timestamp": timestamp,
                        "account": account,
                        "msg_type": msg_type,
                        "qualifier": qualifier,
                        "code": int(code),
                        "code_description": CODES[code]["name"],
                        "category": self._get_event_category(int(code)),
                        "partition": partition,
                        "checksum": checksum,
                        "extension": extension,
                    }

                    self._set_sensor_or_user(event_data, int(sensor_or_user))

                    logger.debug("Event data: %s", event_data)
                    return event_data

        logger.warning("No events found in file %s", path)

    def move_file(self, src: Path) -> None:
        """ Move event file to processed folder """

        dst = self.dst_dir / src.name
        logger.debug("Moving file %s to %s", src, dst)
        src.rename(dst)

    def process_files(self) -> List[Dict[str, Any]]:
        """
        Parse all event files available in spool directory.
        Move each parsed file to another directory
        """
        results = []
        for file in self.src_dir.glob("event-*"):
            if file.is_file():
                event_data = self.parse_file(file)
                if event_data:
                    results.append(event_data)
                if self.move_files:
                    self.move_file(file)
        return results

    @staticmethod
    def _get_uid_from_filename(filename: str) -> str:
        """ Extract unique ID from filename """

        # e.g. event-1IkVo1 -> 1IkVo1
        return filename.replace("event-", "")

    @staticmethod
    def _parse_timestamp_str(timestamp: str) -> int:
        """ Convert the string timestamp coming from Asterisk into seconds from epoch"""
        # e.g
        # Sat Dec 26, 2020 @ 16:16:29 UTC => datetime.datetime(2020, 12, 26, 16, 16, 29)
        return int(datetime.datetime.strptime(timestamp, "%a %b %d, %Y @ %H:%M:%S %Z").timestamp())

    def _set_sensor_or_user(self, event_data: Dict, sensor_or_user: int) -> None:
        """ Set either sensor or user fields """

        # The Ademco standard reuses the 6th field for either sensor or user identification.
        # We look at what each code data type is and set the fields accordingly
        code = str(event_data["code"])
        data_type = CODES[code]["type"]
        if data_type == "zone":
            event_data["sensor"] = sensor_or_user
            # If there are configured sensor names, include the name
            try:
                event_data["sensor_name"] = self.cfg.get("sensors", str(sensor_or_user))
            except (KeyError, configparser.NoOptionError):
                logger.debug("Sensor %s not found in config", str(sensor_or_user))
                event_data["sensor_name"] = None
            event_data["user"] = None
        elif data_type == "user":
            event_data["user"] = sensor_or_user
            event_data["sensor"] = None
            event_data["sensor_name"] = None
        else:
            raise ValueError(f"Invalid data type {data_type}")

    @staticmethod
    def _get_event_category(code: int) -> str:
        """ Given a code number, get the ADEMCO category description """

        bases = [int(s) for s in sorted(EVENT_CATEGORIES)]
        last_base = None
        for b in bases:
            if code >= b:
                last_base = b
            elif code < b:
                break
        return EVENT_CATEGORIES[str(last_base)]



import json
from pathlib import Path

import pytest

from simon_says import events as under_test
from simon_says.ademco import EVENT_CATEGORIES
from simon_says.helpers import redis_present

CWD = Path(__file__).parent
TEST_DATA_DIR = CWD / "data"


def test_process_files(test_parsed_events):
    records = test_parsed_events
    assert len(records) == 2

    e1 = records[0]
    assert e1["uid"] == "12abcd"
    assert e1["account"] == "1234"
    assert e1["extension"] == "simonxt"
    assert e1["code"] == 601
    assert e1["code_description"] == "Manual trigger test report"
    assert e1["category"] == "Test/Misc"
    assert e1["sensor"] == 0
    assert e1["sensor_name"] == "nothing"

    e2 = records[1]
    assert e2["uid"] == "34efgh"
    assert e2["code"] == 131
    assert e2["code_description"] == "Perimeter"
    assert e2["category"] == "Alarms"
    assert e2["sensor"] == 15
    assert e2["sensor_name"] == "front window left"


def test_event(test_parsed_events):
    for r in test_parsed_events:
        event = under_test.AlarmEvent(**r)
        assert isinstance(event, under_test.AlarmEvent)
        assert event.sensor in [0, 15]
        assert json.loads(event.to_json())


@pytest.mark.skipif(not redis_present(), reason="redis not present")
def test_event_store(test_parsed_events, test_db):

    # Create a new event store for testing
    event_store = under_test.EventStore(db=test_db)

    for r in test_parsed_events:
        # Delete it first if it exists from previous tests
        o = event_store.get(r["uid"])
        if o:
            event_store.delete(r["uid"])

        # Create the object
        event = under_test.AlarmEvent(**r)

        # Add it to store
        event_store.add(event)

        # Retrieve it
        e = event_store.get(r["uid"])
        assert e.uid == r["uid"]

    # Get all events
    events = event_store.get_events()
    assert len(events) == 2

    # Clean up
    for r in test_parsed_events:
        event_store.delete(r["uid"])


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (123, EVENT_CATEGORIES["100"]),
        (234, EVENT_CATEGORIES["200"]),
        (345, EVENT_CATEGORIES["300"]),
        (456, EVENT_CATEGORIES["400"]),
        (567, EVENT_CATEGORIES["500"]),
        (678, EVENT_CATEGORIES["600"]),
    ],
)
def test_event_category(test_input, expected):
    assert under_test.EventParser._get_event_category(test_input) == expected



import json
import logging

import falcon

from simon_says.control import Controller
from simon_says.db import DataStore
from simon_says.events import AlarmEvent, EventStore
from simon_says.log import configure_logging
from simon_says.sensors import Sensors, SensorState
from simon_says.version import __version__

logger = logging.getLogger(__name__)


class EventsResource:
    """ API resource for Events """

    def __init__(self, event_store: EventStore, sensors: Sensors, controller: Controller) -> None:

        self.event_store = event_store
        self.sensors = sensors
        self.controller = controller

    def _set_sensor_state(self, event: AlarmEvent) -> None:
        """ Set the sensor state depending on the event code """

        category = event.category
        if event.sensor and category == "Troubles":
            sensor = self.sensors.by_number(event.sensor)
            logger.info("Setting sensor %s (%s) state to 'open'", event.sensor, event.sensor_name)
            sensor.state = SensorState.OPEN
        else:
            logger.debug("_set_sensor_state: Ignoring event %s", event.uid)

    def on_get(self, req, resp, uid: str = None):
        """ Handle GET requests for events in the queue """

        if uid:
            logger.info("Getting event with uid %s", uid)

            try:
                e = self.event_store.get(uid=uid)
                resp.body = e.to_json()
            except KeyError:
                logger.error("uid %s not found", uid)
                raise falcon.HTTPNotFound()

        else:
            logger.info("Getting all events")
            resp.body = self.event_store.events_as_json()

        resp.content_type = "application/json"
        resp.status = falcon.HTTP_200

    def on_post(self, req, resp):
        """ Handle POST requests for event """

        data = req.media
        try:
            logger.info("Adding new event with uid %s", data["uid"])
            event = AlarmEvent(**data)
            self.event_store.add(event)
            self._set_sensor_state(event)
        except Exception as err:
            logger.error("Error creating AlarmEvent: %s", err)
            raise falcon.HTTPBadRequest()

        resp.status = falcon.HTTP_201
        resp.content_type = "application/json"
        resp.body = json.dumps({"result": "OK"})


class ControllerResource:
    """ API resource for commands and state """

    def __init__(self, sensors: Sensors, controller: Controller = None) -> None:
        self.sensors = sensors
        self.controller = controller

    def on_post(self, req, resp):
        """ Handle POST requests for commands """

        data = req.media
        if "action" not in data:
            logger.error("Missing required parameter: 'action'")
            raise falcon.HTTPBadRequest()

        action = data["action"]
        try:
            logger.info("Sending command %s", action)
            if action == "disarm":
                self.controller.disarm()
                self.sensors.clear_all()
            elif action == "arm_home":
                self.controller.arm_home()
            elif action == "arm_away":
                self.controller.arm_away()
            else:
                # Pass other less common actions
                self.controller.send_command(action)

        except Exception as err:
            logger.error("Error sending action to Alarm: %s", err)
            raise falcon.HTTPBadRequest()

        resp.status = falcon.HTTP_202
        resp.content_type = "application/json"
        resp.body = json.dumps({"result": "OK"})


class SensorsResource:
    """ Sensors resource class """

    def __init__(self, sensors: Sensors) -> None:
        self.sensors = sensors

    def on_get(self, req, resp, number: int = None):
        """ Handle GET requests for a given sensor number """

        if number:
            try:
                sensor = self.sensors.by_number(int(number))
                resp.body = sensor.to_json()
            except KeyError:
                logger.error("number %s not found", number)
                raise falcon.HTTPNotFound()
        else:
            logger.info("Getting all sensors")
            resp.body = self.sensors.all_as_json()

        resp.content_type = "application/json"
        resp.status = falcon.HTTP_200


class VersionResource:
    def on_get(self, req, resp):
        """ Handle GET requests for API version """

        resp.content_type = "application/json"
        resp.status = falcon.HTTP_200
        resp.body = json.dumps({"version": __version__})


def create_app(controller: Controller = None, log_level: str = "INFO") -> falcon.API:
    """ Create a Falcon.API object """

    # Wire up the app handler with gunicorn's
    gunicorn_logger = logging.getLogger("gunicorn.error")
    configure_logging(log_level=log_level, handlers=gunicorn_logger.handlers)

    api = falcon.API()

    if not controller:
        controller = Controller()

    version_resource = VersionResource()
    api.add_route("/version", version_resource)

    sensors = Sensors()
    sensors_resource = SensorsResource(sensors=sensors)
    api.add_route("/sensors", sensors_resource)
    api.add_route("/sensors/{number}", sensors_resource)

    db = DataStore()
    events_resource = EventsResource(event_store=EventStore(db=db), sensors=sensors, controller=controller)
    api.add_route("/events", events_resource)
    api.add_route("/events/{uid}", events_resource)

    controller_resource = ControllerResource(sensors=sensors, controller=controller)
    api.add_route("/control", controller_resource)

    return api
